﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Threading;

using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;
using BenchmarkDotNet.Diagnosers;

namespace LockingOverhead {

	// Uncomment the following attribute to tell BenchmarkDotNet to track memory consuption as well (note that input data is allocated outside of benchmarks, thus will not be counted as benchmarks memory overhead.
	// [MemoryDiagnoser]
	public class LockTests {
		[Params(/*10_000,*/ 2_000_000 /*, 20_000_000*/)]	// Uncomment the 20_000_000 to get bigger input set - however the benchmark then can take an hour to complete (depending on the target machine of course).
		public int Files = 2_000_000;   // Pre-assigned so that we can test the class outside of the BenchmarkDotNet infrastructure (can call VerifyCorrectness method).

		public int MaxLength = 128; // 100;

		public const int LockRanges = 16;

		public const int ItemPartitions = 32;

		private FakeString[] filenames;
		private int[] counts;

		[GlobalSetup]
		public void PrepareEnvironment() {
			Random r = new Random();

			filenames = new FakeString[Files];
			for (int i = 0; i < filenames.Length; i++) {
				filenames[i] = new FakeString('X', r.Next(MaxLength - 1) + 1);
			}

			counts = new int[MaxLength];

			globalLock = new object();

			globalMutex = new Mutex();

			itemLocks = new object[MaxLength];
			for (int i = 0; i < itemLocks.Length; i++) {
				itemLocks[i] = new object();
			}

			itemMutexes = new Mutex[MaxLength];
			for (int i = 0; i < itemMutexes.Length; i++) {
				itemMutexes[i] = new Mutex();
			}

			rangeLocks = new object[LockRanges];
			for (int i = 0; i < rangeLocks.Length; i++) {
				rangeLocks[i] = new object();
			}
		}

		[Benchmark(Baseline = true)]
		public void SingleThreaded() {
			Array.Clear(counts, 0, counts.Length);

			foreach (var name in filenames) {
				counts[name.Length]++;
			}
		}

		private object globalLock;

		[Benchmark]
		public void MultiThreadedGlobalLock() {
			Array.Clear(counts, 0, counts.Length);

			Parallel.For(0, filenames.Length, 
				(int i) => {
					int length = filenames[i].Length;
					lock (globalLock) {
						counts[length]++;
					}
				}
			);
		}

		private Mutex globalMutex;

		[Benchmark]
		public void MultiThreadedGlobalMutex_1_100() {
			Array.Clear(counts, 0, counts.Length);

			Parallel.For(0, filenames.Length / 100,
				(int i) => {
					int length = filenames[i].Length;
					globalMutex.WaitOne();
					{
						counts[length]++;
					}
					globalMutex.ReleaseMutex();
				}
			);
		}

		private object[] itemLocks;

		[Benchmark]
		public void MultiThreadedOneLockPerItem() {
			Array.Clear(counts, 0, counts.Length);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					lock (itemLocks[length]) {
						counts[length]++;
					}
				}
			);
		}

		private Mutex[] itemMutexes;

		[Benchmark]
		public void MultiThreadedOneMutexPerItem_1_100() {
			Array.Clear(counts, 0, counts.Length);

			Parallel.For(0, filenames.Length / 100,
				(int i) => {
					int length = filenames[i].Length;
					itemMutexes[length].WaitOne();
					{
						counts[length]++;
					}
					itemMutexes[length].ReleaseMutex();
				}
			);
		}

		private object[] rangeLocks;

		[Benchmark]
		public void MultiThreadedOneLockPerItemRange() {
			Array.Clear(counts, 0, counts.Length);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					lock (rangeLocks[length % LockRanges]) {
						counts[length]++;
					}
				}
			);
		}

		[Benchmark]
		public void MultiThreadedLocalCountsViaThreadLocal() {
			Array.Clear(counts, 0, counts.Length);

			ThreadLocal<int[]> localCounts = new ThreadLocal<int[]>(() => new int[MaxLength], trackAllValues: true);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					localCounts.Value[length]++;
				}
			);

			foreach (var c in localCounts.Values) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		[Benchmark]
		public void MultiThreadedPartitionCountsViaThreadLocalIds() {
			Array.Clear(counts, 0, counts.Length);

			List<int[]> partitionCounts = new List<int[]>();
			object partitionLock = new object();

			ThreadLocal<int> partitionIds = new ThreadLocal<int>(
				() => {
					lock (partitionLock) {
						partitionCounts.Add(new int[MaxLength]);
						int localId = partitionCounts.Count - 1;
						return localId;
					}
				}, trackAllValues: false
			);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					// NOTE: There is a potencial race condition here - calling the partitionCounts[...] indexer getter might race with the partitionCounts.Add(...) call
					// above if a new thread accesses the partitionCounts List, and it needs to "resize" the internal array. However the most probable implementation will
					// replace private pointer to the new array only after is has been properly initialized, and as we never modify the stored value to int[], the race
					// condition should not occur (and as write to a reference variable is atomic, we will here get either the old or the new array of int[], thus it should be OK).
					// However designing thread-safety around undocumented implementation details of another class is a VERY BAD idea!!!
					//
					// SOLUTION? Either lock the partitionLock around partitionCounts getter (which will get the solution even slower) or preallocate a sufficiently sized array [But how big? What if Parallel.For uses more
					// threads than expected?].
					partitionCounts[partitionIds.Value][length]++;
				}
			);

			// Console.WriteLine($"Parallel.For used {partitionCounts.Count} unique threads.");

			foreach (var c in partitionCounts) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		[Benchmark]
		public void MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaLocking() {
			Array.Clear(counts, 0, counts.Length);

			List<int[]> partitionCounts = new List<int[]>();
			object partitionLock = new object();

			ThreadLocal<int> partitionIds = new ThreadLocal<int>(
				() => {
					lock (partitionLock) {
						partitionCounts.Add(new int[MaxLength]);
						int localId = partitionCounts.Count - 1;
						return localId;
					}
				}, trackAllValues: false
			);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					int[] localArray;
					lock (partitionLock) {
						localArray = partitionCounts[partitionIds.Value];
					}
					localArray[length]++;
				}
			);

			foreach (var c in partitionCounts) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		[Benchmark]
		public void MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaPreallocation() {
			Array.Clear(counts, 0, counts.Length);

			int[][] partitionCounts = new int[ItemPartitions][];
			int partitions = 0;

			ThreadLocal<int> partitionIds = new ThreadLocal<int>(
				() => {
					int localId = partitions++;
					partitionCounts[localId] = new int[MaxLength];
					return localId;
				}, trackAllValues: false
			);

			Parallel.For(0, filenames.Length,
				(int i) => {
					int length = filenames[i].Length;
					partitionCounts[partitionIds.Value][length]++;
				}
			);

			foreach (var c in partitionCounts) {
				if (c != null) {
					for (int i = 0; i < c.Length; i++) {
						counts[i] += c[i];
					}
				} else {
					// Skip unused partitions.
				}
			}
		}

		[Benchmark]
		public void MultiThreadedLocalCountsViaPartitions() {
			Array.Clear(counts, 0, counts.Length);

			int[][] partitionCounts = new int[ItemPartitions][];

			Parallel.For(0, ItemPartitions,
				(int partition) => {
					var localCounts = new int[MaxLength];

					int itemsPerPartition = filenames.Length / ItemPartitions;
					int fromIndex = partition * itemsPerPartition;
					int toIndex;
					if (partition == ItemPartitions - 1) {
						toIndex = filenames.Length - 1;
					} else {
						toIndex = (partition + 1) * itemsPerPartition - 1;
					}

					for (int i = fromIndex; i <= toIndex; i++) {
						int length = filenames[i].Length;
						localCounts[length]++;
					}

					partitionCounts[partition] = localCounts;
				}
			);

			foreach (var c in partitionCounts) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		[Benchmark]
		public void AggressiveForLocalCountsViaPartitions() {
			Array.Clear(counts, 0, counts.Length);

			int[][] partitionCounts = new int[ItemPartitions][];

			AggressiveParallelFor(0, ItemPartitions,
				(int partition) => {
					var localCounts = new int[MaxLength];

					int itemsPerPartition = filenames.Length / ItemPartitions;
					int fromIndex = partition * itemsPerPartition;
					int toIndex;
					if (partition == ItemPartitions - 1) {
						toIndex = filenames.Length - 1;
					} else {
						toIndex = (partition + 1) * itemsPerPartition - 1;
					}

					for (int i = fromIndex; i <= toIndex; i++) {
						int length = filenames[i].Length;
						localCounts[length]++;
					}

					partitionCounts[partition] = localCounts;
				}
			);

			foreach (var c in partitionCounts) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		[Benchmark]
		public void VeryAggressiveForLocalCountsViaPartitions() {
			Array.Clear(counts, 0, counts.Length);

			const int manyItemPartitions = ItemPartitions * 16;

			int[][] partitionCounts = new int[manyItemPartitions][];

			AggressiveParallelFor(0, manyItemPartitions,
				(int partition) => {
					var localCounts = new int[MaxLength];

					int itemsPerPartition = filenames.Length / manyItemPartitions;
					int fromIndex = partition * itemsPerPartition;
					int toIndex;
					if (partition == manyItemPartitions - 1) {
						toIndex = filenames.Length - 1;
					} else {
						toIndex = (partition + 1) * itemsPerPartition - 1;
					}

					for (int i = fromIndex; i <= toIndex; i++) {
						int length = filenames[i].Length;
						localCounts[length]++;
					}

					partitionCounts[partition] = localCounts;
				}
			);

			foreach (var c in partitionCounts) {
				for (int i = 0; i < c.Length; i++) {
					counts[i] += c[i];
				}
			}
		}

		private void AggressiveParallelFor(int fromInclusive, int toExclusive, Action<int> body) {
			var threads = new Thread[toExclusive - fromInclusive];

			for (int i = 0; i < threads.Length; i++) {
				int localValue = i;
				threads[i] = new Thread(() => body(localValue));
				threads[i].Start();
			}

			foreach (var t in threads) {
				t.Join();
			}
		}

		public void VerifyCorrectness() {
			PrepareEnvironment();
			var referenceCounts = counts;
			SingleThreaded();
			counts = new int[referenceCounts.Length];

			MultiThreadedGlobalLock();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedGlobalLock));

			MultiThreadedGlobalMutex_1_100();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedGlobalMutex_1_100));

			MultiThreadedOneLockPerItem();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedOneLockPerItem));

			MultiThreadedOneMutexPerItem_1_100();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedOneMutexPerItem_1_100));

			MultiThreadedOneLockPerItemRange();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedOneLockPerItemRange));

			MultiThreadedLocalCountsViaThreadLocal();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedLocalCountsViaThreadLocal));

			MultiThreadedPartitionCountsViaThreadLocalIds();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedPartitionCountsViaThreadLocalIds));

			MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaLocking();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaLocking));

			MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaPreallocation();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedPartitionCountsViaThreadLocalIdsCorrectedViaPreallocation));

			MultiThreadedLocalCountsViaPartitions();
			CheckIfCountsCorrect(referenceCounts, nameof(MultiThreadedLocalCountsViaPartitions));

			AggressiveForLocalCountsViaPartitions();
			CheckIfCountsCorrect(referenceCounts, nameof(AggressiveForLocalCountsViaPartitions));

			VeryAggressiveForLocalCountsViaPartitions();
			CheckIfCountsCorrect(referenceCounts, nameof(VeryAggressiveForLocalCountsViaPartitions));
		}

		private void CheckIfCountsCorrect(int[] referenceCounts, string testName) {
			Console.WriteLine(testName);
			for (int i = 0; i < counts.Length; i++) {
				if (counts[i] != referenceCounts[i]) {
					Console.WriteLine($"[{i}] == {counts[i]} should be {referenceCounts[i]}");
					break;
				}
			}
			Console.WriteLine("DONE.");
		}


		// We are using fake strings with just .Length API to save memory so the benchmark can run on a reasonable machine with less than 4 GiB of RAM.
		class FakeString {
			public FakeString(char c, int count) {
				Length = count;
			}

			public int Length { get; }
		}
	}

	class Program {
		static void Main(string[] args) {
			// Uncomment the following code to test correctness of the implemented algorithms.
			//var tests = new LockTests();
			//tests.VerifyCorrectness();
			
			BenchmarkRunner.Run<LockTests>();
		}
	}
}
